import asyncio
from typing import NoReturn
import os
import sys
from appdirs import user_config_dir


GLOW_INSTALLATION_URL = (
    "https://github.com/charmbracelet/glow"
)
APP_NAME = "Sengpt"
APP_NAME_LOWER = "sengpt"
REPO_URL = "https://github.com/SenZmaKi/Sengpt"
DESCRIPTION = "ChatGPT in your terminal, no OpenAI API key required"
REPO_TAGS_URL = "https://api.github.com/repos/SenZmaKi/Sengpt/tags"
# FIXME: Update to False on push
DEBUG = False
ROOT_DIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
VERSION = "0.1.0"
V_VERSION = f"v{VERSION}"


class OsUtils:
    is_windows = sys.platform == "win32"
    is_linux = sys.platform == "linux"
    is_mac = sys.platform == "mac"
    config_dir = ROOT_DIR if DEBUG else user_config_dir(APP_NAME)
    if is_windows:
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


def print_and_exit(text: str) -> NoReturn:
    print(text)
    sys.exit()


def check_repo_print(text: str, exit=True) -> None | NoReturn:
    check_repo_text = f"{text}\nCheck {REPO_URL} for help"
    if exit:
        print_and_exit(check_repo_text)
    print(check_repo_text)


def mkdir(path: str) -> None:
    # Incase some weirdo has a file with the same name as the folder
    if os.path.isfile(path):
        os.unlink(path)
    if not os.path.isdir(path):
        os.mkdir(path)





import sys

from sengpt.argparser import SYS_ARGS
from .utils import APP_NAME, mkdir, OsUtils, check_repo_print
from appdirs import user_config_dir
import os
from typing import Any, NoReturn, TypeVar
import json

T = TypeVar("T")


class Config:
    @staticmethod
    def update_session_token():
        if len(SYS_ARGS.non_args) != 1:
            check_repo_print(
                "Invalid usage!!!\nUsage: --session_token <your-session-token-goes-here>"
            )
        session_token = SYS_ARGS.non_args[0]
        Config.session_token = session_token
        Config.update_json("session_token", session_token)

    @staticmethod
    def setup_config_file_path() -> str:
        config_dir = user_config_dir(APP_NAME)
        if OsUtils.is_windows:
            # On windows it resolves the config directory to Sengpt/Sengpt
            config_dir = os.path.dirname(config_dir)
        mkdir(config_dir)
        return os.path.join(config_dir, "config.json")

    @staticmethod
    def info() -> str:
        with open(Config.file_path, "r") as f:
            return f"{f.read()}\n\nConfig file location: {Config.file_path}"

    @staticmethod
    def update_json(key: str, value: str | None | bool) -> None:
        Config.json[key] = value
        Config.dump()

    @staticmethod
    async def update_json_async(key: str, value: str | None | bool) -> None:
        Config.update_json(key, value)

    @staticmethod
    def dump() -> None:
        with open(Config.file_path, "w") as f:
            json.dump(Config.json, f, indent=4)

    @staticmethod
    def get_from_json_config(
        key: str, default: T, json_config: dict[str, Any]
    ) -> T | NoReturn:
        value = json_config.get(key, default)
        val_type = type(value)
        def_type = type(default)
        if val_type == def_type:
            return value
        print(
            f'Error loading config: Expected value at "{key}" to be of type {def_type} but instead got "{value}" which is of type {val_type}'
        )
        sys.exit()

    @staticmethod
    def load_json_config(
        config_file_path: str,
    ) -> NoReturn | dict[str, Any]:
        try:
            with open(config_file_path) as f:
                contents = f.read()
                try:
                    return json.loads(contents)
                except json.JSONDecodeError:
                    last_2_chars = contents.replace(" ", "").replace("\n", "")[-2:]
                    info_to_add = (
                        '\nRemove the "," before the last "}"'
                        if ",}" == last_2_chars
                        else ""
                    )
                    check_repo_print(
                        f'Your config file at: "{config_file_path}" is invalid!!!{info_to_add}'
                    )
        except FileNotFoundError:
            pass
        return {}  # This is here instead of in the except block to avoid type errors

    file_path = setup_config_file_path()

    json = load_json_config(file_path)
    username = get_from_json_config("username", "You", json)
    session_token = get_from_json_config("session_token", "", json)
    model = get_from_json_config("model", "gpt-3.5", json)
    recent_conversation_id = get_from_json_config("recent_conversation_id", "", json)
    preconfigured_prompts = get_from_json_config("preconfigured_prompts", {}, json)
    no_glow = get_from_json_config("no_glow", False, json)
    save = get_from_json_config("save", False, json)
    delete = get_from_json_config("delete", False, json)
    copy = get_from_json_config("copy", False, json)
    default_mode = get_from_json_config("default_mode", "interactive", json)



import ctypes
import inspect
import time
import uuid
import websockets
from websockets.exceptions import ConnectionClosed
import json
import base64
import asyncio
from queue import Queue
from threading import Thread
from typing import Callable, Generator, Optional

from curl_cffi.requests import Session

from .async_chatgpt import (
    BACKUP_ARKOSE_TOKEN_GENERATOR,
    CHATGPT_API,
    USER_AGENT,
    AsyncChatGPT,
    AsyncConversation,
    MODELS,
    WS_REGISTER_URL,
)
from .errors import (
    BackendError,
    InvalidSessionToken,
    RetryError,
    TokenNotProvided,
    UnexpectedResponseError,
    InvalidModelName,
)
from .utils import sync_get_binary_path, get_model_slug


class SyncConversation(AsyncConversation):
    def __init__(self, chatgpt, conversation_id: Optional[str] = None, model=None):
        super().__init__(chatgpt, conversation_id, model)

    def fetch_chat(self) -> dict:
        """
        Fetches the chat of the conversation from the API.

        Returns:
            dict: The JSON response from the API containing the chat if the conversation_id is not none, else returns an empty dict.

        Raises:
            UnexpectedResponseError: If the response is not a valid JSON object or if the response json is not in the expected format
        """
        if not self.conversation_id:
            return {}

        url = CHATGPT_API.format(f"conversation/{self.conversation_id}")
        response = self.chatgpt.session.get(
            url=url, headers=self.chatgpt.build_request_headers()
        )

        error = None
        try:
            chat = response.json()
            self.parent_id = list(chat.get("mapping", {}))[-1]
            model_slug = get_model_slug(chat)
            self.model = [
                key for key, value in MODELS.items() if value["slug"] == model_slug
            ][0]
        except Exception as e:
            error = e
        if error is not None:
            raise UnexpectedResponseError(error, response.text)

        return chat

    def chat(self, user_input: str) -> Generator[dict, None, None]:
        """
        As the name implies, chat with ChatGPT.

        Args:
            user_input (str): The user's input message.

        Yields:
            dict: A dictionary representing assistant responses.

        Returns:
            Generator[dict, None]: A generator object that yields assistant responses.

        Raises:
            UnexpectedResponseError: If the response is not a valid JSON object or if the response json is not in the expected format
        """

        payload = self.build_message_payload(user_input)

        server_response = (
            ""  # To store what the server returned for debugging in case of an error
        )
        error = None
        try:
            full_message = None
            while True:
                response = self.send_message(payload=payload) if not self.chatgpt.websocket_mode else self.send_websocket_message(payload=payload)
                for chunk in response:
                    decoded_chunk = chunk.decode() if not self.chatgpt.websocket_mode else chunk

                    server_response += decoded_chunk
                    for line in decoded_chunk.splitlines():
                        if not line.startswith("data: "):
                            continue

                        raw_json_data = line[6:]
                        if not (decoded_json := self.decode_raw_json(raw_json_data)):
                            continue

                        if (
                            "message" in decoded_json
                            and decoded_json["message"]["author"]["role"] == "assistant"
                        ):
                            processed_response = self.filter_response(decoded_json)
                            if full_message:
                                prev_resp_len = len(
                                    full_message["message"]["content"]["parts"][0]
                                )
                                processed_response["content"] = processed_response[
                                    "content"
                                ][prev_resp_len::]

                            yield processed_response
                            full_message = decoded_json
                self.conversation_id = full_message["conversation_id"]
                self.parent_id = full_message["message"]["id"]
                if (
                    full_message["message"]["metadata"]["finish_details"]["type"]
                    == "max_tokens"
                ):
                    payload = self.build_message_continuation_payload()
                else:
                    break
        except Exception as e:
            error = e

        # raising the error outside the 'except' block to prevent the 'During handling of the above exception, another exception occurred' error
        if error is not None:
            raise UnexpectedResponseError(error, server_response)

    def send_message(self, payload: dict) -> Generator[bytes, None, None]:
        """
        Send a message payload to the server and receive the response.

        Args:
            payload (dict): Payload containing message information.

        Yields:
            bytes: Chunk of data received as a response.
        """
        response_queue = Queue()

        def perform_request():
            def content_callback(chunk):
                response_queue.put(chunk)

            url = CHATGPT_API.format("conversation")
            headers = self.chatgpt.build_request_headers()
            # Add Chat Requirements Token
            chat_requriments_token = self.chatgpt.create_chat_requirements_token()
            if chat_requriments_token:
                headers["openai-sentinel-chat-requirements-token"] = chat_requriments_token

            response = self.chatgpt.session.post(
                url=url,
                headers=headers,
                json=payload,
                content_callback=content_callback,
            )
            response_queue.put(None)

        Thread(target=perform_request).start()

        while True:
            chunk = response_queue.get()
            if chunk is None:
                break
            yield chunk
    
    def send_websocket_message(self, payload: dict) -> Generator[str, None, None]:
        """
        Send a message payload via WebSocket and receive the response.

        Args:
            payload (dict): Payload containing message information.

        Yields:
            str: Chunk of data received as a response.
        """

        response_queue = Queue()
        websocket_request_id = None

        def perform_request():
            nonlocal websocket_request_id
            
            url = CHATGPT_API.format("conversation")
            headers = self.chatgpt.build_request_headers()
            # Add Chat Requirements Token
            chat_requriments_token = self.chatgpt.create_chat_requirements_token()
            if chat_requriments_token:
                headers["openai-sentinel-chat-requirements-token"] = chat_requriments_token

            response = (self.chatgpt.session.post(
                url=url,
                headers=headers,
                json=payload,
            )).json()

            websocket_request_id = response.get("websocket_request_id")
            
            if websocket_request_id is None:
                raise UnexpectedResponseError("WebSocket request ID not found in response", response)

            if websocket_request_id not in self.chatgpt.ws_conversation_map:
                self.chatgpt.ws_conversation_map[websocket_request_id] = response_queue
            
        Thread(target=perform_request).start()

        while True:
            chunk = response_queue.get()
            if chunk is None:
                break
            yield chunk

        del self.chatgpt.ws_conversation_map[websocket_request_id]

    def build_message_payload(self, user_input: str) -> dict:
        """
        Build a payload for sending a user message.

        Returns:
            dict: Payload containing message information.
        """
        if self.conversation_id and (self.parent_id is None or self.model is None):
            self.fetch_chat()  # it will automatically fetch the chat and set the parent id

        payload = {
            "conversation_mode": {"conversation_mode": {"kind": "primary_assistant"}},
            "conversation_id": self.conversation_id,
            "action": "next",
            "arkose_token": self.arkose_token_generator()
            if self.chatgpt.generate_arkose_token
            or MODELS[self.model]["needs_arkose_token"]
            else None,
            "force_paragen": False,
            "history_and_training_disabled": False,
            "messages": [
                {
                    "author": {"role": "user"},
                    "content": {"content_type": "text", "parts": [user_input]},
                    "id": str(uuid.uuid4()),
                    "metadata": {},
                }
            ],
            "model": MODELS[self.model]["slug"],
            "parent_message_id": str(uuid.uuid4())
            if not self.parent_id
            else self.parent_id,
            "websocket_request_id": str(uuid.uuid4())
            if self.chatgpt.websocket_mode
            else None,
        }

        return payload

    def build_message_continuation_payload(self) -> dict:
        """
        Build a payload for continuing ChatGPT's cut off response.

        Returns:
            dict: Payload containing message information for continuation.
        """
        payload = {
            "conversation_mode": {"conversation_mode": {"kind": "primary_assistant"}},
            "action": "continue",
            "arkose_token": self.arkose_token_generator()
            if self.chatgpt.generate_arkose_token
            or MODELS[self.model]["needs_arkose_token"]
            else None,
            "conversation_id": self.conversation_id,
            "force_paragen": False,
            "history_and_training_disabled": False,
            "model": MODELS[self.model]["slug"],
            "parent_message_id": self.parent_id,
            "timezone_offset_min": -300,
        }

        return payload

    def arkose_token_generator(self) -> str:
        """
        Generate an Arkose token.

        Returns:
            str: Arkose token.
        """
        if not self.chatgpt.tried_downloading_binary:
            self.chatgpt.binary_path = sync_get_binary_path(self.chatgpt.session)

            if self.chatgpt.binary_path:
                self.chatgpt.arkose = ctypes.CDLL(self.chatgpt.binary_path)
                self.chatgpt.arkose.GetToken.restype = ctypes.c_char_p

            self.chatgpt.tried_downloading_binary = True

        if self.chatgpt.binary_path:
            try:
                result = self.chatgpt.arkose.GetToken()
                return ctypes.string_at(result).decode("utf-8")
            except:
                pass

        for _ in range(5):
            response = self.chatgpt.session.get(BACKUP_ARKOSE_TOKEN_GENERATOR)
            if response.text == "null":
                raise BackendError(error_code=505)
            try:
                return response.json()["token"]
            except:
                time.sleep(0.7)

        raise RetryError(website=BACKUP_ARKOSE_TOKEN_GENERATOR)

    def delete(self) -> None:
        """
        Deletes the conversation.
        """
        if self.conversation_id:
            self.chatgpt.delete_conversation(self.conversation_id)

            self.conversation_id = None
            self.parent_id = None


class SyncChatGPT(AsyncChatGPT):
    def __init__(
        self,
        proxies: Optional[dict] = None,
        session_token: Optional[str] = None,
        exit_callback_function: Optional[Callable] = None,
        auth_token: Optional[str] = None,
        websocket_mode: Optional[bool] = False,
    ):
        """
        Initializes an instance of the class.

        Args:
            proxies (Optional[dict]): A dictionary of proxy settings. Defaults to None.
            session_token (Optional[str]): A session token. Defaults to None.
            exit_callback_function (Optional[callable]): A function to be called on exit. Defaults to None.
            auth_token (Optional[str]): An authentication token. Defaults to None.
            websocket_mode (Optional[bool]): Toggle whether to use WebSocket for chat. Defaults to False.
        """
        super().__init__(
            proxies=proxies,
            session_token=session_token,
            exit_callback_function=exit_callback_function,
            auth_token=auth_token,
            websocket_mode=websocket_mode,
        )

        self.stop_websocket_flag = False
        self.stop_websocket = None

    def __enter__(self):
        self.session = Session(
            impersonate="chrome110", timeout=99999, proxies=self.proxies
        )

        if self.generate_arkose_token:
            self.binary_path = sync_get_binary_path(self.session)

            if self.binary_path:
                self.arkose = ctypes.CDLL(self.binary_path)
                self.arkose.GetToken.restype = ctypes.c_char_p

            self.tried_downloading_binary = True

        if not self.auth_token:
            if self.session_token is None:
                raise TokenNotProvided
            self.auth_token = self.fetch_auth_token()
            
        # automaticly check the status of websocket_mode
        if not self.websocket_mode:
            self.websocket_mode = self.check_websocket_availability()
            
        if self.websocket_mode:
            def run_websocket():
                asyncio.run(self.ensure_websocket())
            self.ws_loop = Thread(target=run_websocket)
            self.ws_loop.start()

        return self

    def __exit__(self, *args):
        try:
            if self.exit_callback_function and callable(self.exit_callback_function):
                if not inspect.iscoroutinefunction(self.exit_callback_function):
                    self.exit_callback_function(self)
        finally:
            self.session.close()

        if self.websocket_mode:
            self.stop_websocket_flag = True
            self.ws_loop.join()

    def get_conversation(self, conversation_id: str) -> SyncConversation:
        """
        Makes an instance of class Conversation and return it.

        Args:
            conversation_id (str): The ID of the conversation to fetch.

        Returns:
            Conversation: Conversation object.
        """

        return SyncConversation(self, conversation_id)

    def create_new_conversation(
        self, model: Optional[str] = "gpt-3.5"
    ) -> SyncConversation:
        if model not in MODELS:
            raise InvalidModelName(model, MODELS)
        return SyncConversation(self, model=model)

    def delete_conversation(self, conversation_id: str) -> dict:
        """
        Delete a conversation.

        Args:
            conversation_id (str): Unique identifier for the conversation.

        Returns:
            dict: Server response json.
        """
        url = CHATGPT_API.format(f"conversation/{conversation_id}")
        response = self.session.patch(
            url=url, headers=self.build_request_headers(), json={"is_visible": False}
        )

        return response.json()

    def fetch_auth_token(self) -> str:
        """
        Fetch the authentication token for the session.

        Raises:
            InvalidSessionToken: If the session token is invalid.

        Returns: authentication token.
        """
        url = "https://chat.openai.com/api/auth/session"
        cookies = {"__Secure-next-auth.session-token": self.session_token}

        headers = {
            "User-Agent": USER_AGENT,
            "Accept": "*/*",
            "Accept-Language": "en-US,en;q=0.5",
            "Alt-Used": "chat.openai.com",
            "Connection": "keep-alive",
            "Sec-Fetch-Dest": "empty",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Site": "same-origin",
            "Sec-GPC": "1",
            "Cookie": "; ".join(
                [
                    f"{cookie_key}={cookie_value}"
                    for cookie_key, cookie_value in cookies.items()
                ]
            ),
        }

        response = self.session.get(url=url, headers=headers)
        response_json = response.json()

        if "accessToken" in response_json:
            return response_json["accessToken"]

        raise InvalidSessionToken

    def set_custom_instructions(
        self,
        about_user: Optional[str] = "",
        about_model: Optional[str] = "",
        enable_for_new_chats: Optional[bool] = True,
    ) -> dict:
        """
        Set cuteom instructions for ChatGPT.

        Args:
            about_user (str): What would you like ChatGPT to know about you to provide better responses?
            about_model (str): How would you like ChatGPT to respond?
            enable_for_new_chats (bool): Enable for new chats.
        Returns:
            dict: Server response json.
        """
        data = {
            "about_user_message": about_user,
            "about_model_message": about_model,
            "enabled": enable_for_new_chats,
        }
        url = CHATGPT_API.format("user_system_messages")
        response = self.session.post(
            url=url, headers=self.build_request_headers(), json=data
        )

        return response.json()

    def retrieve_chats(
        self, offset: Optional[int] = 0, limit: Optional[int] = 28
    ) -> dict:
        params = {
            "offset": offset,
            "limit": limit,
            "order": "updated",
        }
        url = CHATGPT_API.format("conversations")
        response = self.session.get(
            url=url, params=params, headers=self.build_request_headers()
        )

        return response.json()
    
    def check_websocket_availability(self) -> bool:
        """
        Check if WebSocket is available.

        Returns:
            bool: True if WebSocket is available, otherwise False.
        """
        url = CHATGPT_API.format("accounts/check/v4-2023-04-27")
        response = (self.session.get(
            url=url, headers=self.build_request_headers()
        )).json()
        
        if 'account_ordering' in response and 'accounts' in response:
            account_id = response['account_ordering'][0]
            if account_id in response['accounts']:
                return 'shared_websocket' in response['accounts'][account_id]['features']

        return False
    
    async def ensure_websocket(self):
        ws_url_rsp = self.session.post(WS_REGISTER_URL, headers=self.build_request_headers()).json()
        ws_url = ws_url_rsp['wss_url']
        access_token = self.extract_access_token(ws_url)
        asyncio.create_task(self.ensure_close_websocket())
        await self.listen_to_websocket(ws_url, access_token)
        
    async def ensure_close_websocket(self):
        while True:
            if self.stop_websocket_flag:
                break
            await asyncio.sleep(1)
        await self.stop_websocket()

    async def listen_to_websocket(self, ws_url: str, access_token: str):
        headers = {'Authorization': f'Bearer {access_token}'}
        async with websockets.connect(ws_url, extra_headers=headers) as websocket:
            async def stop_websocket():
                await websocket.close()
            self.stop_websocket = stop_websocket

            while True:
                message = None
                try:
                    message = await websocket.recv()
                except ConnectionClosed:
                    break
                message_data = json.loads(message)
                body_encoded = message_data.get("body", "")
                ws_id = message_data.get("websocket_request_id", "")
                decoded_body = base64.b64decode(body_encoded).decode('utf-8')
                response_queue = self.ws_conversation_map.get(ws_id)
                if response_queue is None:
                    continue
                response_queue.put_nowait(decoded_body)
                if '[DONE]' in decoded_body or '[ERROR]' in decoded_body:
                    response_queue.put(None)
                    continue

    def create_chat_requirements_token(self):
        """
        Get a chat requirements token from chatgpt server

        Returns:
            str: chat requirements token
        """
        url = CHATGPT_API.format("sentinel/chat-requirements")
        response = self.session.post(
            url=url, headers=self.build_request_headers()
        )
        body = response.json()
        token = body.get("token", None)
        return token


import asyncio
from typing import NoReturn, cast
from curl_cffi.requests.errors import RequestsError
from curl_cffi.requests.session import AsyncSession
from .re_gpt import AsyncChatGPT
import sys
import subprocess

from .re_gpt.async_chatgpt import AsyncConversation
from .re_gpt.errors import UnexpectedResponseError
from .re_gpt.sync_chatgpt import InvalidSessionToken
from .utils import (
    REPO_TAGS_URL,
    GLOW_INSTALLATION_URL,
    V_VERSION,
    OsUtils,
    check_repo_print,
    print_and_exit,
)
from .config import Config
from .argparser import ArgParser, SYS_ARGS
import pyperclip
import os

PRINT_WITH_GLOW = not (SYS_ARGS.is_set("no_glow") or Config.no_glow)
# isatty == is a teletypewriter == is a terminal == program invoked without piping input
INPUT_WAS_PIPED = not sys.stdin.isatty()
IS_QUERY_MODE = (
    SYS_ARGS.is_set("query")
    # Couldn't find a way to get piped inputs to work together with normal user inputs
    # cause stdin gets set to the piped input and but I couldn't find a way to set it to the terminal stdin
    or INPUT_WAS_PIPED
    or (Config.default_mode == "query" and not SYS_ARGS.is_set("interactive"))
)


def input_handler(msg: str) -> str:
    try:
        user_input = input(msg)
    except EOFError:  # When they press Ctrl + C while input is "waiting"
        print()  # Move out of the msg line onto the next line
        user_input = ""
    return user_input


def till_one_works(commands: tuple[str, ...]) -> bool:
    for c in commands:
        try:
            subprocess.run(c)
            return True
        except Exception:  # I can't be asked to catch all the specific errors that can arise from these commands
            continue
    return False


def try_installing_glow() -> bool:
    if OsUtils.is_windows:
        return till_one_works(
            (
                "winget install charmbracelet.glow",
                "choco install glow",
                "scoop install glow",
            )
        )
    elif OsUtils.is_linux:
        return till_one_works(
            (
                'sudo mkdir -p /etc/apt/keyrings && curl -fsSL https://repo.charm.sh/apt/gpg.key | sudo gpg --dearmor -o /etc/apt/keyrings/charm.gpg && echo "deb [signed-by=/etc/apt/keyrings/charm.gpg] https://repo.charm.sh/apt/ * *" | sudo tee /etc/apt/sources.list.d/charm.list && sudo apt update && sudo apt install glow',
                "pacman -S glow",
                "xbps-install -S glow",
                "nix-env -iA nixpkgs.glow",
                "pkg install glow",
                "eopkg install glow",
            )
        )
    return till_one_works(("brew install glow", "sudo port install glow"))


def get_piped_input() -> str:
    if not INPUT_WAS_PIPED:
        return ""
    full_input = ""
    try:
        while True:
            full_input = f"{full_input}\n{input()}"
    except EOFError:
        pass
    return full_input


def generate_prompt(args: ArgParser) -> str:
    args_prompt = " ".join(args.non_args)
    if args_prompt:
        args_prompt = f"{args_prompt}"
    preconfigured_prompt = " ".join(
        value for key, value in Config.preconfigured_prompts.items() if args.is_set(key)
    )
    if preconfigured_prompt:
        preconfigured_prompt = f"{preconfigured_prompt}\n\n"
    clipboard_text = f"{pyperclip.paste()}" if args.is_set("paste") else ""
    if clipboard_text:
        clipboard_text = f"{clipboard_text}\n\n"
    passed_input = get_piped_input() if IS_QUERY_MODE else ""
    passed_input = f"{passed_input}\n\n" if passed_input else ""
    return f"{passed_input}{clipboard_text}{preconfigured_prompt}{args_prompt}"


async def loading_animation(event: asyncio.Event) -> None:
    animation = (".   ", "..  ", ".. .", "    ")
    while not event.is_set():
        for a in animation:
            print(f"  Thinking {a}", end="\r")
            await asyncio.sleep(0.1)
    print(" " * (len("Thinking") + 8), end="\r")


def printer(text: str) -> None:
    if PRINT_WITH_GLOW:
        return glow_print(text)
    print(text)


def glow_print(text: str) -> None:
    try:
        subprocess.run("glow", input=text.encode())
    except FileNotFoundError:
        print("Glow is required to pretty print output!!!")
        user_choice = input_handler("Would you like to install it? [y/n]\n> ").lower()
        if user_choice == "y" or user_choice == "yes":
            installation_was_succesful = try_installing_glow()
            if installation_was_succesful:
                return glow_print(text)
            print("Automatic installation failed X(")
        print(
            f'Check {GLOW_INSTALLATION_URL} for an installation guide\nAlternatively you can set "no_glow" to true in your config file or pass the --no_glow flag to the program'
        )
        os._exit(
            1
        )  # os._exit instead of sys.exit to avoid asyncio errors leaking cause running tasks don't respect sys.exit


def load_conversation(gpt: AsyncChatGPT) -> tuple[AsyncConversation, bool]:
    conversation_id = (
        Config.recent_conversation_id
        if SYS_ARGS.is_set("recent_conversation")
        else None
    )
    if conversation_id == "":
        conversation_id = None

    conversation = gpt.create_new_conversation(model=Config.model)
    conversation.conversation_id = conversation_id
    save_conversation = (
        conversation_id is not None or Config.save or SYS_ARGS.is_set("save")
    )
    return conversation, save_conversation


async def fetch_prompt_response(prompt: str, conversation: AsyncConversation) -> str:
    prompt_response = ""
    event = asyncio.Event()
    loading_task = asyncio.create_task(loading_animation(event))
    async for resp_json in conversation.chat(prompt):
        content = resp_json["content"]
        if PRINT_WITH_GLOW:
            prompt_response += content
            continue
        if not event.is_set():
            event.set()
            await loading_task
        print(content, end="", flush=True)
    if PRINT_WITH_GLOW:
        event.set()
        await loading_task
    return prompt_response


async def interactive_mode(
    args: ArgParser,
    conversation: AsyncConversation,
    is_first_iteration=True,
) -> None:
    if is_first_iteration:
        prompt = prepare_prompt(args)
    else:
        prompt = generate_prompt(args)
        printer("\n# ChatGPT")
    prompt_response = await fetch_prompt_response(prompt, conversation)
    printer(f"{prompt_response}\n\n# {Config.username}")
    handle_coping_to_clip(args, prompt_response)
    if user_input := input_handler("> "):
        if user_input == "-d" or user_input == "--delete":
            await conversation.delete()
            return
        await interactive_mode(ArgParser(user_input.split(" ")), conversation)
        return
    if Config.delete:
        await conversation.delete()
        return
    await Config.update_json_async(
        "recent_conversation_id", conversation.conversation_id
    )


def handle_coping_to_clip(args: ArgParser, prompt_response: str) -> None:
    if args.is_set("copy") or Config.copy:
        pyperclip.copy(prompt_response)


def prepare_prompt(args: ArgParser) -> str:
    prompt = generate_prompt(args)
    preprompt_text = f"# {Config.username}\n{prompt}\n\n# ChatGPT\n"
    printer(preprompt_text)
    return prompt


async def query_mode(
    args: ArgParser, conversation: AsyncConversation, save_conversation: bool
) -> None:
    prompt = prepare_prompt(args)
    prompt_response = await fetch_prompt_response(prompt, conversation)
    if save_conversation:
        Config.recent_conversation_id = cast(str, conversation.conversation_id)
        task = asyncio.create_task(
            Config.update_json_async(
                "recent_conversation_id", conversation.conversation_id
            )
        )
    else:
        task = asyncio.create_task(conversation.delete())
    printer(prompt_response)
    handle_coping_to_clip(args, prompt_response)
    await task


async def gpt_coroutine(gpt: AsyncChatGPT) -> None:
    try:
        async with gpt:
            conversation, save_conversation = load_conversation(gpt)
            if IS_QUERY_MODE:
                await query_mode(SYS_ARGS, conversation, save_conversation)
                return
            await interactive_mode(SYS_ARGS, conversation)

    except (UnexpectedResponseError, InvalidSessionToken) as e:
        if isinstance(e, asyncio.CancelledError):
            return
        elif isinstance(e, InvalidSessionToken):
            print_and_exit("Invalid session token, make a new one")
        elif "token_expired" in e.message:
            check_repo_print("Your session token has expired, make a new one")
        else:
            raise


async def update_check_coroutine(session: AsyncSession) -> bool:
    response = await session.get(REPO_TAGS_URL)
    tags = response.json()
    # Incase I delete all tags for whatever reason or the repo gets taken down
    if not tags or not isinstance(tags, list):
        return False
    latest_tag = tags[0]
    if latest_tag["name"] != V_VERSION:
        return True
    return False


async def async_main() -> None:
    gpt = AsyncChatGPT(session_token=Config.session_token)
    session = AsyncSession(impersonate="chrome110")
    try:
        _, update_is_available = await asyncio.gather(
            gpt_coroutine(gpt), update_check_coroutine(session)
        )
        if update_is_available:
            print('\n\nUpdate available run "pip update sengpt" to install it')
    except RequestsError:
        print("Check your internet!!!")


def validate_session_token() -> None | NoReturn:
    if not Config.session_token:
        check_repo_print("Session token must be provided during initial configuration")



def handle_static_args() -> None | NoReturn:
    if SYS_ARGS.is_set("version"):
        print(ArgParser.version_info())
        sys.exit()
    if SYS_ARGS.is_set("help"):
        print(ArgParser.help_info())
        sys.exit()
    if SYS_ARGS.is_set("config_file"):
        print(Config.info())
        sys.exit()
    if SYS_ARGS.is_set("session_token"):
        Config.update_session_token()
        print("\nSuccessfully set session token")
        sys.exit()


def main():
    handle_static_args()
    validate_session_token()
    try:
        asyncio.run(async_main())
    except KeyboardInterrupt:
        return


if __name__ == "__main__":
    main()



import sys

from .utils import APP_NAME_LOWER, DESCRIPTION, VERSION


class ArgParser:
    def __init__(self, args: list[str]) -> None:
        self.args, self.non_args = ArgParser.parse_args(args)

    @staticmethod
    def parse_args(args: list[str]) -> tuple[list[str], list[str]]:
        non_args: list[str] = []
        extracted_args: list[str] = []
        for a in args:
            if a.startswith("-"):
                extracted_args.append(a)
                continue
            non_args.append(a)
        return extracted_args, non_args

    @staticmethod
    def version_info() -> str:
        return f"{APP_NAME_LOWER} {VERSION}"

    @staticmethod
    def help_info() -> str:
        return f"""
{DESCRIPTION}

Usage: senpgt [prompt] [options]

-h, --help                    Show help message and exit                                                
-v, --version                 Show the version information                                              
-cf, --config_file            Show the config file's contents and location                              
-st, --session_token          Set session token                                                         
                                                                                                        
-ng, --no_glow                Disable pretty printing with Glow,                                        
                              this can be set to be the default behaviour in the config file                
                                                                                                        
-c,  --copy                   Copy the prompt response to the clipboard,                                
                              this can be set to be the default behaviour in the config file                
                                                                                                        
-p,  --paste                  Append the most recently copied clipboard text to the sent prompt         
-rc, --recent_conversation    Use the most recently saved conversation as context                       
-pp, --preconfigured_prompt   Append a preconfigured prompt to the sent prompt,                         
                              replace "preconfigured_prompt" with the prompt's name                     
                              as it appears in the config file                                          
                                                                                                        
-q, query                     Use query mode i.e., print ChatGPT's response and exit,                   
                              this flag is only necessary if "default_mode" in config file is interactive 
                                                                                                        
-s, --save                    By default conversations in query mode are deleted on exit,               
                              this saves the conversation instead,                                      
                              this can be set to be the default behaviour in the config file                
                                                                                                        
-i, --interactive             Use interactive mode i.e., back and forth interaction with ChatGPT,       
                              this flag is only necessary if "default_mode" in the config file is query   
                                                                                                        
-d, --delete                  By default conversations in interactive mode are saved on exit,           
                              this deletes then exits the interactive mode session,                     
                              this can be set to be the default behaviour in the config file                
"""

    @staticmethod
    def short_and_long(name: str) -> tuple[str, str]:
        long = f"--{name}"
        chars: list[str] = []
        for prev_char, char in zip(name, name[1:]):
            if prev_char == "_":
                chars.append(char)
        short = f'-{name[0]}{"".join(chars)}'
        return short, long

    @staticmethod
    def abstract_is_set(flag_name: str, args: list[str]) -> bool:
        short, long = ArgParser.short_and_long(flag_name)
        return short in args or long in args

    def is_set(self, flag_name: str) -> bool:
        return ArgParser.abstract_is_set(flag_name, self.args)


SYS_ARGS = ArgParser(sys.argv[1:])




import asyncio
import ctypes
import inspect
import json
import uuid
import re
import websockets
import base64
from typing import AsyncGenerator, Callable, Optional

from curl_cffi.requests import AsyncSession
from .errors import (
    BackendError,
    InvalidSessionToken,
    RetryError,
    TokenNotProvided,
    UnexpectedResponseError,
    InvalidModelName,
)
from .utils import async_get_binary_path, get_model_slug

# Constants
USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/110.0.0.0 Safari/537.36"
CHATGPT_API = "https://chat.openai.com/backend-api/{}"
BACKUP_ARKOSE_TOKEN_GENERATOR = "https://arkose-token-generator.zaieem.repl.co/token"
WS_REGISTER_URL = CHATGPT_API.format("register-websocket")

MODELS = {
    "gpt-4": {"slug": "gpt-4", "needs_arkose_token": True},
    "gpt-3.5": {"slug": "text-davinci-002-render-sha", "needs_arkose_token": False},
}


class AsyncConversation:
    def __init__(self, chatgpt, conversation_id=None, model=None):
        self.chatgpt = chatgpt
        self.conversation_id = conversation_id
        self.parent_id = None
        self.model = model

    async def fetch_chat(self) -> dict:
        """
        Fetches the chat of the conversation from the API.

        Returns:
            dict: The JSON response from the API containing the chat if the conversation_id is not none, else returns an empty dict.

        Raises:
            UnexpectedResponseError: If the response is not a valid JSON object or if the response json is not in the expected format
        """
        if not self.conversation_id:
            return {}

        url = CHATGPT_API.format(f"conversation/{self.conversation_id}")
        response = await self.chatgpt.session.get(
            url=url, headers=self.chatgpt.build_request_headers()
        )

        error = None
        try:
            chat = response.json()
            self.parent_id = list(chat.get("mapping", {}))[-1]
            model_slug = get_model_slug(chat)
            self.model = [
                key for key, value in MODELS.items() if value["slug"] == model_slug
            ][0]
        except Exception as e:
            error = e
        if error is not None:
            raise UnexpectedResponseError(error, response.text)

        return chat

    async def chat(self, user_input: str) -> AsyncGenerator[dict, None]:
        """
        As the name implies, chat with ChatGPT.

        Args:
            user_input (str): The user's input message.

        Yields:
            dict: A dictionary representing assistant responses.

        Returns:
            AsyncGenerator[dict, None]: An asynchronous generator object that yields assistant responses.

        Raises:
            UnexpectedResponseError: If the response is not a valid JSON object or if the response json is not in the expected format
        """

        payload = await self.build_message_payload(user_input)

        server_response = (
            ""  # To store what the server returned for debugging in case of an error
        )
        error = None
        try:
            full_message = None
            while True:
                response = self.send_message(payload=payload) if not self.chatgpt.websocket_mode else self.send_websocket_message(payload=payload)
                async for chunk in response:
                    decoded_chunk = chunk.decode() if isinstance(chunk, bytes) else chunk
                    
                    server_response += decoded_chunk
                    for line in decoded_chunk.splitlines():
                        if not line.startswith("data: "):
                            continue

                        raw_json_data = line[6:]
                        if not (decoded_json := self.decode_raw_json(raw_json_data)):
                            continue

                        if (
                            "message" in decoded_json
                            and decoded_json["message"]["author"]["role"] == "assistant"
                        ):
                            processed_response = self.filter_response(decoded_json)
                            if full_message:
                                prev_resp_len = len(
                                    full_message["message"]["content"]["parts"][0]
                                )
                                processed_response["content"] = processed_response[
                                    "content"
                                ][prev_resp_len::]

                            yield processed_response
                            full_message = decoded_json
                self.conversation_id = full_message["conversation_id"]
                self.parent_id = full_message["message"]["id"]
                if (
                    full_message["message"]["metadata"]["finish_details"]["type"]
                    == "max_tokens"
                ):
                    payload = await self.build_message_continuation_payload()
                else:
                    break
        except Exception as e:
            error = e

        # raising the error outside the 'except' block to prevent the 'During handling of the above exception, another exception occurred' error
        if error is not None:
            raise UnexpectedResponseError(error, server_response)

    async def send_message(self, payload: dict) -> AsyncGenerator[bytes, None]:
        """
        Send a message payload to the server and receive the response.

        Args:
            payload (dict): Payload containing message information.

        Yields:
            bytes: Chunk of data received as a response.
        """
        response_queue = asyncio.Queue()

        async def perform_request():
            def content_callback(chunk):
                response_queue.put_nowait(chunk)

            url = CHATGPT_API.format("conversation")
            
            headers = self.chatgpt.build_request_headers()
            # Add Chat Requirements Token
            chat_requriments_token = await self.chatgpt.create_chat_requirements_token()
            if chat_requriments_token:
                headers["openai-sentinel-chat-requirements-token"] = chat_requriments_token

            await self.chatgpt.session.post(
                url=url,
                headers=headers,
                json=payload,
                content_callback=content_callback,
            )
            await response_queue.put(None)

        asyncio.create_task(perform_request())

        while True:
            chunk = await response_queue.get()
            if chunk is None:
                break
            yield chunk
    
    async def send_websocket_message(self, payload: dict) -> AsyncGenerator[str, None]:
        """
        Send a message payload via WebSocket and receive the response.

        Args:
            payload (dict): Payload containing message information.

        Yields:
            str: Chunk of data received as a response.
        """
        await self.chatgpt.ensure_websocket()

        response_queue = asyncio.Queue()
        websocket_request_id = None

        async def perform_request():
            nonlocal websocket_request_id
            
            url = CHATGPT_API.format("conversation")
            headers = self.chatgpt.build_request_headers()
            # Add Chat Requirements Token
            chat_requriments_token = await self.chatgpt.create_chat_requirements_token()
            if chat_requriments_token:
                headers["openai-sentinel-chat-requirements-token"] = chat_requriments_token

            response = (await self.chatgpt.session.post(
                url=url,
                headers=headers,
                json=payload,
            )).json()

            websocket_request_id = response.get("websocket_request_id")
            
            if websocket_request_id is None:
                raise UnexpectedResponseError("WebSocket request ID not found in response", response)
            
            if websocket_request_id not in self.chatgpt.ws_conversation_map:
                self.chatgpt.ws_conversation_map[websocket_request_id] = response_queue
            
        asyncio.create_task(perform_request())

        while True:
            chunk = await response_queue.get()
            if chunk is None:
                break
            yield chunk

        del self.chatgpt.ws_conversation_map[websocket_request_id]
    

    async def build_message_payload(self, user_input: str) -> dict:
        """
        Build a payload for sending a user message.

        Returns:
            dict: Payload containing message information.
        """
        if self.conversation_id and (self.parent_id is None or self.model is None):
            await self.fetch_chat()  # it will automatically fetch the chat and set the parent id

        payload = {
            "conversation_mode": {"conversation_mode": {"kind": "primary_assistant"}},
            "conversation_id": self.conversation_id,
            "action": "next",
            "arkose_token": await self.arkose_token_generator()
            if self.chatgpt.generate_arkose_token
            or MODELS[self.model]["needs_arkose_token"]
            else None,
            "force_paragen": False,
            "history_and_training_disabled": False,
            "messages": [
                {
                    "author": {"role": "user"},
                    "content": {"content_type": "text", "parts": [user_input]},
                    "id": str(uuid.uuid4()),
                    "metadata": {},
                }
            ],
            "model": MODELS[self.model]["slug"],
            "parent_message_id": str(uuid.uuid4())
            if not self.parent_id
            else self.parent_id,
            "websocket_request_id": str(uuid.uuid4())
            if self.chatgpt.websocket_mode
            else None,
        }

        return payload

    async def build_message_continuation_payload(self) -> dict:
        """
        Build a payload for continuing ChatGPT's cut off response.

        Returns:
            dict: Payload containing message information for continuation.
        """
        payload = {
            "conversation_mode": {"conversation_mode": {"kind": "primary_assistant"}},
            "action": "continue",
            "arkose_token": await self.arkose_token_generator()
            if self.chatgpt.generate_arkose_token
            or MODELS[self.model]["needs_arkose_token"]
            else None,
            "conversation_id": self.conversation_id,
            "force_paragen": False,
            "history_and_training_disabled": False,
            "model": MODELS[self.model]["slug"],
            "parent_message_id": self.parent_id,
            "timezone_offset_min": -300,
        }

        return payload

    async def arkose_token_generator(self) -> str:
        """
        Generate an Arkose token.

        Returns:
            str: Arkose token.
        """
        if not self.chatgpt.tried_downloading_binary:
            self.chatgpt.binary_path = await async_get_binary_path(self.chatgpt.session)

            if self.chatgpt.binary_path:
                self.chatgpt.arkose = ctypes.CDLL(self.chatgpt.binary_path)
                self.chatgpt.arkose.GetToken.restype = ctypes.c_char_p

            self.chatgpt.tried_downloading_binary = True

        if self.chatgpt.binary_path:
            try:
                result = self.chatgpt.arkose.GetToken()
                return ctypes.string_at(result).decode("utf-8")
            except:
                pass

        for _ in range(5):
            response = await self.chatgpt.session.get(BACKUP_ARKOSE_TOKEN_GENERATOR)
            if response.text == "null":
                raise BackendError(error_code=505)
            try:
                return response.json()["token"]
            except:
                await asyncio.sleep(0.7)

        raise RetryError(website=BACKUP_ARKOSE_TOKEN_GENERATOR)

    async def delete(self) -> None:
        """
        Deletes the conversation.
        """
        if self.conversation_id:
            await self.chatgpt.delete_conversation(self.conversation_id)

            self.conversation_id = None
            self.parent_id = None

    @staticmethod
    def decode_raw_json(raw_json_data: str) -> dict or bool:
        """
        Decode JSON.

        Args:
            raw_json_data (str): JSON as a string.

        Returns:
            dict: Decoded JSON.
        """
        try:
            decoded_json = json.loads(raw_json_data.strip())
            return decoded_json
        except:
            return False

    @staticmethod
    def filter_response(response):
        processed_response = {
            "content": response["message"]["content"]["parts"][0],
            "message_id": response["message"]["id"],
            "parent_id": response["message"]["metadata"]["parent_id"],
            "conversation_id": response["conversation_id"],
        }

        return processed_response


class AsyncChatGPT:
    def __init__(
        self,
        proxies: Optional[dict] = None,
        session_token: Optional[str] = None,
        exit_callback_function: Optional[Callable] = None,
        auth_token: Optional[str] = None,
        generate_arkose_token: Optional[bool] = False,
        websocket_mode: Optional[bool] = False,
    ):
        """
        Initializes an instance of the class.

        Args:
            proxies (Optional[dict]): A dictionary of proxy settings. Defaults to None.
            session_token (Optional[str]): A session token. Defaults to None.
            exit_callback_function (Optional[callable]): A function to be called on exit. Defaults to None.
            auth_token (Optional[str]): An authentication token. Defaults to None.
            generate_arkose_token (Optional[bool]): Toggle whether to generate and send arkose-token in the payload. Defaults to False.
            websocket_mode (Optional[bool]): Toggle whether to use WebSocket for chat. Defaults to False.
        """
        self.proxies = proxies
        self.exit_callback_function = exit_callback_function

        self.arkose = None
        self.binary_path = None
        self.tried_downloading_binary = False
        self.generate_arkose_token = generate_arkose_token

        self.session_token = session_token
        self.auth_token = auth_token
        self.session = None
        
        self.websocket_mode = websocket_mode
        self.ws_loop = None
        self.ws_conversation_map = {}

    async def __aenter__(self):
        self.session = AsyncSession(
            impersonate="chrome110", timeout=99999, proxies=self.proxies
        )
        if self.generate_arkose_token:
            self.binary_path = await async_get_binary_path(self.session)

            if self.binary_path:
                self.arkose = ctypes.CDLL(self.binary_path)
                self.arkose.GetToken.restype = ctypes.c_char_p

            self.tried_downloading_binary = True

        if not self.auth_token:
            if self.session_token is None:
                raise TokenNotProvided
            self.auth_token = await self.fetch_auth_token()

        if not self.websocket_mode:
           self.websocket_mode = await self.check_websocket_availability()

        if self.websocket_mode:
            await self.ensure_websocket()

        return self

    async def __aexit__(self, *_):
        try:
            if self.exit_callback_function and callable(self.exit_callback_function):
                if not inspect.iscoroutinefunction(self.exit_callback_function):
                    self.exit_callback_function(self)
        finally:
            self.session.close()

    def build_request_headers(self) -> dict:
        """
        Build headers for HTTP requests.

        Returns:
            dict: Request headers.
        """
        headers = {
            "User-Agent": USER_AGENT,
            "Accept": "text/event-stream",
            "Accept-Language": "en-US",
            "Accept-Encoding": "gzip, deflate, br",
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.auth_token}",
            "Origin": "https://chat.openai.com",
            "Alt-Used": "chat.openai.com",
            "Connection": "keep-alive",
        }

        return headers

    def get_conversation(self, conversation_id: str) -> AsyncConversation:
        """
        Makes an instance of class Conversation and return it.

        Args:
            conversation_id (str): The ID of the conversation to fetch.

        Returns:
            Conversation: Conversation object.
        """

        return AsyncConversation(self, conversation_id)

    def create_new_conversation(
        self, model: Optional[str] = "gpt-3.5"
    ) -> AsyncConversation:
        if model not in MODELS:
            raise InvalidModelName(model, MODELS)
        return AsyncConversation(self, model=model)

    async def delete_conversation(self, conversation_id: str) -> dict:
        """
        Delete a conversation.

        Args:
            conversation_id (str): Unique identifier for the conversation.

        Returns:
            dict: Server response json.
        """
        url = CHATGPT_API.format(f"conversation/{conversation_id}")
        response = await self.session.patch(
            url=url, headers=self.build_request_headers(), json={"is_visible": False}
        )

        return response.json()

    async def fetch_auth_token(self) -> str:
        """
        Fetch the authentication token for the session.

        Raises:
            InvalidSessionToken: If the session token is invalid.

        Returns: authentication token.
        """
        url = "https://chat.openai.com/api/auth/session"
        cookies = {"__Secure-next-auth.session-token": self.session_token}

        headers = {
            "User-Agent": USER_AGENT,
            "Accept": "*/*",
            "Accept-Language": "en-US,en;q=0.5",
            "Alt-Used": "chat.openai.com",
            "Connection": "keep-alive",
            "Sec-Fetch-Dest": "empty",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Site": "same-origin",
            "Sec-GPC": "1",
            "Cookie": "; ".join(
                [
                    f"{cookie_key}={cookie_value}"
                    for cookie_key, cookie_value in cookies.items()
                ]
            ),
        }

        response = await self.session.get(url=url, headers=headers)
        response_json = response.json()

        if "accessToken" in response_json:
            return response_json["accessToken"]

        raise InvalidSessionToken

    async def set_custom_instructions(
        self,
        about_user: Optional[str] = "",
        about_model: Optional[str] = "",
        enable_for_new_chats: Optional[bool] = True,
    ) -> dict:
        """
        Set cuteom instructions for ChatGPT.

        Args:
            about_user (str): What would you like ChatGPT to know about you to provide better responses?
            about_model (str): How would you like ChatGPT to respond?
            enable_for_new_chats (bool): Enable for new chats.
        Returns:
            dict: Server response json.
        """
        data = {
            "about_user_message": about_user,
            "about_model_message": about_model,
            "enabled": enable_for_new_chats,
        }
        url = CHATGPT_API.format("user_system_messages")
        response = await self.session.post(
            url=url, headers=self.build_request_headers(), json=data
        )

        return response.json()

    async def retrieve_chats(
        self, offset: Optional[int] = 0, limit: Optional[int] = 28
    ) -> dict:
        params = {
            "offset": offset,
            "limit": limit,
            "order": "updated",
        }
        url = CHATGPT_API.format("conversations")
        response = await self.session.get(
            url=url, params=params, headers=self.build_request_headers()
        )

        return response.json()

    async def check_websocket_availability(self) -> bool:
        """
        Check if WebSocket is available.

        Returns:
            bool: True if WebSocket is available, otherwise False.
        """
        url = CHATGPT_API.format("accounts/check/v4-2023-04-27")
        response = (await self.session.get(
            url=url, headers=self.build_request_headers()
        )).json()
        
        if 'account_ordering' in response and 'accounts' in response:
            account_id = response['account_ordering'][0]
            if account_id in response['accounts']:
                return 'shared_websocket' in response['accounts'][account_id]['features']

        return False
    
    async def ensure_websocket(self):
        if not self.ws_loop:
            ws_url_rsp = (await self.session.post(WS_REGISTER_URL, headers=self.build_request_headers())).json()
            ws_url = ws_url_rsp['wss_url']
            access_token = self.extract_access_token(ws_url)
            self.ws_loop = asyncio.create_task(self.listen_to_websocket(ws_url, access_token))

    def extract_access_token(self, url):
        match = re.search(r'access_token=([^&]*)', url)
        if match:
            return match.group(1)
        else:
            return None
        
    async def listen_to_websocket(self, ws_url: str, access_token: str):
        headers = {'Authorization': f'Bearer {access_token}'}
        async with websockets.connect(ws_url, extra_headers=headers) as websocket:
            while True:
                message = await websocket.recv()
                message_data = json.loads(message)
                body_encoded = message_data.get("body", "")
                ws_id = message_data.get("websocket_request_id", "")
                decoded_body = base64.b64decode(body_encoded).decode('utf-8')
                response_queue = self.ws_conversation_map.get(ws_id)
                if response_queue is None:
                    continue
                if 'title_generation' in decoded_body:
                    # skip
                    continue
                response_queue.put_nowait(decoded_body)
                if '[DONE]' in decoded_body or '[ERROR]' in decoded_body:
                    await response_queue.put(None)
                    continue

    async def create_chat_requirements_token(self):
        """
        Get a chat requirements token from chatgpt server

        Returns:
            str: chat requirements token
        """
        url = CHATGPT_API.format("sentinel/chat-requirements")
        response = await self.session.post(
            url=url, headers=self.build_request_headers()
        )
        body = response.json()
        token = body.get("token", None)
        return token
